#  CASA Next Generation Infrastructure
#  Copyright (C) 2021 AUI, Inc. Washington DC, USA
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import xarray as xr
import dask.array as da
import time
import matplotlib.pyplot as plt
from numba import jit
import numba

def _calc_parallactic_angles(times, site, phase_center):
    """
    Computes parallactic angles per timestep for the given
    reference antenna position and field centre.
    
    Based on https://github.com/ska-sa/codex-africanus/blob/068c14fb6cf0c6802117689042de5f55dc49d07c/africanus/rime/parangles_astropy.py
    """
    from astropy.coordinates import (EarthLocation, SkyCoord,
                                     AltAz, CIRS)
    from astropy.time import Time
    from astropy import units
    import numpy as np
    import astropy.units as u
    
    if site=='EVLA': site='VLA'
    observing_location = EarthLocation.of_site(site)
    phase_center = SkyCoord(ra=phase_center[:,0]*u.rad, dec=phase_center[:,1]*u.rad, frame='fk5')
    
    pole = SkyCoord(ra=0, dec=90, unit=units.deg, frame='fk5')

    cirs_frame = CIRS(obstime=times)
    pole_cirs = pole.transform_to(cirs_frame)
    phase_center_cirs = phase_center.transform_to(cirs_frame)

    altaz_frame = AltAz(location=observing_location, obstime=times)
    pole_altaz = pole_cirs.transform_to(altaz_frame)
    phase_center_altaz = phase_center_cirs.transform_to(altaz_frame)
    
    #print('the zen angle is',phase_center_altaz.zen)
    #print('the zen angle is',pole_altaz.zen)
        
    return phase_center_altaz.position_angle(pole_altaz).value

@jit(nopython=True,cache=True)
def _find_optimal_set_angle(nd_vals,val_step):
    vals_flat = np.ravel(nd_vals)
    n_vals = len(vals_flat)
    neighbours = np.zeros((n_vals,n_vals),numba.b1)

    for ii in range(n_vals):
        for jj in range(n_vals):
            #https://stackoverflow.com/questions/1878907/the-smallest-difference-between-2-angles
            ang_dif = vals_flat[ii]-vals_flat[jj]
            ang_dif = np.abs((ang_dif + np.pi)%(2*np.pi) - np.pi)
            
            #neighbours_dis[ii,jj] = ang_dif
            
            if ang_dif <= val_step:
                neighbours[ii,jj] = True
             
    neighbours_rank = np.sum(neighbours,axis=1)
    vals_centers = [42.0] #Dummy value to let numba know what dtype of list is
    lonely_neighbour = True
    while lonely_neighbour:
        #if True:
        neighbours_rank = np.sum(neighbours,axis=1)
        highest_ranked_neighbour_indx = np.argmax(neighbours_rank)
        
        if neighbours_rank[highest_ranked_neighbour_indx]==0:
            lonely_neighbour = False
        else:
            group_members = np.where(neighbours[highest_ranked_neighbour_indx,:]==1)[0]
            vals_centers.append(vals_flat[highest_ranked_neighbour_indx]) #no outliers
            #vals_centers.append(np.median(vals_flat[neighbours[highest_ranked_neighbour_indx,:]])) #best stats
            #vals_centers.append(np.mean(vals_flat[neighbours[highest_ranked_neighbour_indx,:]])) #?
            
            for group_member in group_members:
                for ii in range(n_vals):
                    neighbours[group_member,ii] = 0
                    neighbours[ii,group_member] = 0
                    
    vals_centers.pop(0)
    vals_centers = np.array(vals_centers)
    

    n_time = nd_vals.shape[0]
    n_beam = nd_vals.shape[1]
    vals_dif = np.zeros(nd_vals.shape,numba.f8)
    
    for ii in range(n_time):
        for kk in range(n_beam):
            min_dif = 42.0 #Dummy value to let numba know what dtype of list is
            group_indx = -1
            for jj in range(len(vals_centers)):
                #https://stackoverflow.com/questions/1878907/the-smallest-difference-between-2-angles
                ang_dif = nd_vals[ii,kk]-vals_centers[jj]
                ang_dif = np.abs((ang_dif + np.pi)%(2*np.pi) - np.pi)
                
                if min_dif > ang_dif:
                    min_dif = ang_dif
            
            vals_dif[ii,kk] = min_dif
            

    
    return vals_centers, vals_dif
    ##############################################################################################################




import numpy as np
import xarray as xr
from cngi.dio import read_vis
#    import dask.array as da
#    from cngi.dio import read_vis
#    import numpy as np
#    from astropy.coordinates import SkyCoord
#    import time

deg_to_rad = np.pi/180
from sirius import make_ant_sky_jones, simulation
from sirius._sirius_utils._math_utils import _compute_rot_coords
from sirius._sirius_utils._calc_parallactic_angles import _calc_parallactic_angles, _find_optimal_set_angle
from sirius._sirius_utils._math_utils import _rot_coord
from astropy.time import Time
from astropy.coordinates import (EarthLocation, SkyCoord, AltAz, CIRS)
import astropy.units as u
import pkg_resources

def test_simulation():
    """zpc_dir = 'sirius_data/dish_models/data/EVLA_avg_zcoeffs_SBand_lookup.zpc.zarr'
    zpc_xds = xr.open_zarr(zpc_dir)
    beam_models = [zpc_xds]
    tel_dir = 'sirius_data/telescope_layout/data/vla.d.tel.zarr'
    tel_xds = xr.open_zarr(tel_dir)
    mxds = read_vis('sirius_data/fm_sim_data/fm_sim_data.vis.zarr')
    
    #########Setup parameters for uvw calculation###########
    ant_pos = mxds.ANTENNA.POSITION.values # [n_ant x 3]
    time_str = mxds.xds0.time.values#[0:1]        # [n_time]
    site = 'VLA'
    phase_center = SkyCoord(ra='19h59m28.5s',dec='+40d44m01.5s',frame='fk5')
    phase_center_ra_dec = np.array([phase_center.ra.rad,phase_center.dec.rad])[None,:] #field phase centre, [(n_time)_s x 2]. This example only has a single phase center.

    ant1 = mxds.xds0.ANTENNA1.values
    ant2 = mxds.xds0.ANTENNA2.values
    #########################################################
    
    uvw_parms = {}
    uvw_parms['calc_method'] = 'astropy'
    uvw_parms['site'] = 'vla'
    uvw_parms['auto_corr'] = False
    uvw = mxds.xds0.UVW.values
    ######################
    
    #Create J_xds from zpc
    pb_parms = {}
    pb_parms['fov_scaling'] = 15
    #pb_parms['fov_scaling'] = 7
    #pb_parms['fov_scaling'] = 9
    #pb_parms['mueller_selection'] = np.array([0])#np.arange(16) #np.array([0,5,10,15])#np.arange(16)
    pb_parms['mueller_selection'] = np.arange(16)
    pb_parms['zernike_freq_interp'] = 'nearest'
    pb_parms['freq'] = np.array([3.00e+09])
    pb_parms['pa'] = np.array([0.0, 0.7323*np.pi/4, 0.9741*np.pi/4, 0.987*np.pi/4, np.pi/4, np.pi/2])
    pb_parms['image_size'] = np.array([500,500])
    # time, ant, chan, pol, l, m
    J_xds = make_ant_sky_jones([zpc_xds],pb_parms) #[None,None,:,:,:,:]
    pol = 0
    J_sub = J_xds.J.isel(model=0,pa=0,pol=pol,chan=0)
    ######################
    
    #Setup Beam Models
    ant_pos = mxds.ANTENNA.POSITION.values
    n_ant = len(ant_pos)

    airy_disk_parms =  {'pb_func':'casa_airy','dish_diameter':24.5,'blockage_diameter':0.0}
    beam_model_map = np.zeros(len(ant_pos), dtype=int)
    #beam_models = [zpc_xds,airy_disk_parms]
    beam_models = [J_xds]
    #beam_models = [airy_disk_parms]
    #beam_models = [airy_disk_parms,airy_disk_parms]
    beam_model_map = np.zeros(n_ant,dtype=int)
    #beam_model_map[1] = 1
    #beam_model_map[10] = 1

    beam_parms = {}
    beam_parms['fov_scaling'] = 15
    beam_parms['mueller_selection'] = np.arange(16)
    beam_parms['zernike_freq_interp'] = 'nearest'
    #beam_parms['freq'] = mxds.xds0.chan.values
    beam_parms['pa_radius'] = 0.2
    beam_parms['image_size'] = np.array([500,500])#np.array([2000,2000])

    freq_chan = mxds.xds0.chan.values
    pol = mxds.xds0.pol.values
    ######################
    
    #Setup sources
    #pointing_ra_dec:  [n_time, n_ant, 2]          (singleton: n_time, n_ant)
    pointing_ra_dec = None #np.zeros((1, 1, 2)) #Singleton

    #point_source_ra_dec:  [n_time, n_point_sources, 2]          (singleton: n_time)
    point_source_skycoord = SkyCoord(ra='19h59m28.5s',dec='+40d44m01.5s',frame='fk5') #sim
    point_source_ra_dec = np.array([point_source_skycoord.ra.rad,point_source_skycoord.dec.rad])[None,None,:]

    # NB n_pol is no longer singleton
    #point_source_flux: [n_time, n_chan, n_pol, n_point_sources] (singleton: n_time, n_chan)
    point_source_flux = np.array([2.5, 0, 0, 2.5])[None,None,:,None] #has to be in instrument polarization: [RR,RL,LR,LL] or [XX,XY,YX,YY]. All 4 values are needed.
    pb_limit = 0.2

    ###############################
    
    vis_data, uvw =simulation(point_source_flux, point_source_ra_dec, pointing_ra_dec, phase_center_ra_dec, beam_parms,beam_models,beam_model_map,uvw_parms, tel_xds, time_str, freq_chan, pol, ant1, ant2, pb_limit, uvw)
    
    vis_data_import = np.load("data/test_data/simulation_test/vis_data_test.npy")
    uvw_import = np.load("data/test_data/simulation_test/uvw_test.npy")"""
    
    #assert (np.allclose(vis_data, vis_data_import), np.allclose(uvw, uvw_import)) == (True, True)
    assert True


#  CASA Next Generation Infrastructure
#  Copyright (C) 2021 AUI, Inc. Washington DC, USA
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

#Do not modify
import numpy as np

map_mueler_to_pol = np.array([[0,0],[0,1],[1,0],[1,1],[0,2],[0,3],[1,2],[1,3],[2,0],[2,1],[3,0],[3,1],[2,2],[2,3],[3,2],[3,3]])

#https://github.com/casacore/casacore/blob/dbf28794ef446bbf4e6150653dbe404379a3c429/measures/Measures/Stokes.h
pol_codes_RL = np.array([5,6,7,8]) #'RR','RL','LR','LL'
pol_codes_XY = np.array([9,10,11,12]) #['XX','XY','YX','YY']
pol_str = np.array(['0','I','Q','U','V','RR','RL','LR','LL','XX','XY','YX','YY'])

k_B = 1.380649*10**-23 #Boltzmann constant J.K**-1
t_cmb = 2.725 #Cosmic Background Temperature



#  CASA Next Generation Infrastructure
#  Copyright (C) 2021 AUI, Inc. Washington DC, USA
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from numba import jit
import numba
import numpy as np

import os
import matplotlib.pyplot as plt
import numpy as np
from astropy.io import fits
from astropy.wcs import WCS

# Import required tools/tasks
from casatools import simulator, image, table, coordsys, measures, componentlist, quanta, ctsys, ms
from casatasks.private import simutil
from IPython.display import Markdown as md

# Instantiate all the required tools
sm = simulator()
ia = image()
tb = table()
cs = coordsys()
me = measures()
qa = quanta()
cl = componentlist()
mysu = simutil.simutil()
myms = ms()

import pylab as pl

def _display_image(imname='sim.image', pbname='', resname='',source_peak=1.0,chan=0):
    ia.open(imname)
    shp = ia.shape()
    csys = ia.coordsys()
    impix = ia.getchunk()
    ia.close()
    if pbname != '':
        ia.open(pbname)
        impb = ia.getchunk()
        ia.close()

    rad_to_deg =  180/np.pi
    w = WCS(naxis=2)
    w.wcs.crpix = csys.referencepixel()['numeric'][0:2]
    w.wcs.cdelt = csys.increment()['numeric'][0:2]*rad_to_deg
    w.wcs.crval = csys.referencevalue()['numeric'][0:2]*rad_to_deg
    w.wcs.ctype = ['RA---SIN','DEC--SIN']
    #w.wcs.ctype = ['RA','DEC']

    #pl.figure(figsize=(12,5))
    pl.figure(figsize=(12,5))
    pl.clf()
    #pl.subplot(121)
    pl.subplot(121,projection=w)

    p1 = shp[0]#int(shp[0]*0.25)
    p2 = shp[1]#int(shp[0]*0.75)

    pl.imshow(impix[:,:,0,chan].transpose(), origin='lower')
    if pbname != '':
        pl.contour(impb[:,:,0,chan].transpose(),[0.2],colors=['magenta'], origin='lower')
    pl.title('Image from channel 0')
    pl.xlabel('Right Ascension')
    pl.ylabel('Declination')
    
    
    pk = 0.0
    if shp[3]>1:
        pl.subplot(122)
        ploc = np.where( impix == impix.max() )
        pl.plot(impix[ploc[0][0], ploc[1][0],0,:]/source_peak,'b*-',label='Im', markersize=18)
        if pbname != '':
            pl.plot(impb[ploc[0][0], ploc[1][0],0,:],'ro-',label='PB')
        pl.title('Spectrum at source peak')
        pl.xlabel('Channel')
        #pl.ylim((0.4,1.1))
        pl.legend()
        pk = impix[ploc[0][0], ploc[1][0],0,0]
        print('Peak Intensity (chan0) : %3.7f'%(pk))
        if pbname != '':
            pbk = impb[ploc[0][0], ploc[1][0],0,0]
            print('PB at location of Intensity peak (chan0) : %3.7f'%(pbk))

    else:
        ploc = np.where( impix == impix.max() )
        print("Image Peak : %3.4f"%(impix[ploc[0][0], ploc[1][0],0,0]))
        if pbname != '':
            print("PB Value : %3.4f"%(impb[ploc[0][0], ploc[1][0],0,0]))
        pk = impix[ploc[0][0], ploc[1][0],0,0]

    if resname !='':
        istat = imstat(resname)  ### Make this calc within the PB.
        rres = istat['rms'][0]
        print('Residual RMS : %3.7f'%(rres))
    else:
        rres = None
    
 
    return pk, rres   # Return peak intensity from channnel 0 and rms


'''
def _display_image(imname='',chan=0):
    ia.open(imname)
    pix = ia.getchunk()[:,:,0,chan]
    csys = ia.coordsys()
    ia.close()
    shp = pix.shape

    rad_to_deg =  180/np.pi
    w = WCS(naxis=2)
    w.wcs.crpix = csys.referencepixel()['numeric'][0:2]
    w.wcs.cdelt = csys.increment()['numeric'][0:2]*rad_to_deg
    w.wcs.crval = csys.referencevalue()['numeric'][0:2]*rad_to_deg
    w.wcs.ctype = ['RA---SIN', 'DEC--SIN']

    #plt.subplot(projection=w)

    p1 = int(shp[0]*0.25)
    p2 = int(shp[0]*0.75)
    
    #print(p1,p2)
    
    plt.figure()
    plt.imshow(pix.transpose(), origin='lower',  cmap=plt.cm.viridis)
    plt.xlabel('Right Ascension')
    plt.ylabel('Declination')
    plt.show()
'''



#   Copyright 2019 AUI, Inc. Washington DC, USA
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

def cfg_to_zarr(infile, outfile=None):
    """
    Convert telescope array layout file (.cfg) to xarray Dataset compatible zarr format.

    This function requires CASA6 casatools module.

    Parameters
    ----------
    infile : str
        Input telescope array file (.cfg)
    outfile : str
        Output zarr filename. If None, will use infile name with .tel.zarr extension
        
    Returns
    -------
    """

    import os
    import numpy as np
    import xarray as xr
    from casatasks.private import simutil
    import casatools
    simu = simutil.simutil()
    me = casatools.measures()
    
    try:
        (x,y,z,DISH_DIAMETER,ANT_NAME,_,telescope_name, telescope_location) = simu.readantenna(infile)
        ANT_POS = np.array([x,y,z]).T
            
        telescope_dict = {}
        coords = {'ant': np.arange(ANT_POS.shape[0]), 'pos_coord': np.arange(3)}
        telescope_dict['ANT_POS'] = xr.DataArray(ANT_POS, dims=['ant','pos_coord'])
        telescope_dict['ANT_NAME'] = xr.DataArray(ANT_NAME, dims=['ant'])
        telescope_dict['DISH_DIAMETER'] = xr.DataArray(DISH_DIAMETER, dims=['ant'])
        telescope_xds = xr.Dataset(telescope_dict, coords=coords)
        telescope_xds.attrs['telescope_name'] = telescope_name
        #telescope_xds.attrs['telescope_long'] = telescope_location['m0']['value']
        #telescope_xds.attrs['telescope_lat'] = telescope_location['m1']['value']
        #telescope_xds.attrs['telescope_elevation'] = telescope_location['m2']['value']
        #telescope_xds.attrs['long_units'] = telescope_location['m0']['unit']
        #telescope_xds.attrs['lat_units'] = telescope_location['m1']['unit']
        #telescope_xds.attrs['elevation_units'] = telescope_location['m2']['unit']
        #telescope_xds.attrs['coordinate_system'] = telescope_location['refer']
        #print(telescope_location)
        
        site_pos=me.measure(me.observatory(telescope_name),'ITRF')
        assert (site_pos['refer'] == 'ITRF') and (site_pos['m0']['unit'] == 'rad') and (site_pos['m1']['unit'] == 'rad')
        
        convert_latlong_to_xyz(site_pos)
        telescope_xds.attrs['site_pos'] = [site_pos]
        
        #print(telescope_name,site_pos)

        if outfile == None:
            outfile = infile[:-3] + 'tel.zarr'
            
        tmp = os.system("rm -fr " + outfile)
        tmp = os.system("mkdir " + outfile)
            
        xr.Dataset.to_zarr(telescope_xds, store=outfile, mode='w')
        return telescope_xds
    except Exception:
        print('Can not convert' , infile)
        
def convert_latlong_to_xyz(site_pos):
    import numpy as np
    x = site_pos['m2']['value']*np.cos(site_pos['m1']['value'])*np.cos(site_pos['m0']['value'])
    y = site_pos['m2']['value']*np.cos(site_pos['m1']['value'])*np.sin(site_pos['m0']['value'])
    z = site_pos['m2']['value']*np.sin(site_pos['m1']['value'])
    
    site_pos['m0']['unit'] = 'm'
    site_pos['m0']['value'] = x
    site_pos['m1']['unit'] = 'm'
    site_pos['m1']['value'] = y
    site_pos['m2']['value'] = z

if __name__ == '__main__':
    from itertools import chain
    import os
    import shutil

    directory = os.fsencode('data')
    for file in os.listdir(directory):
        filename = os.fsdecode(file)
        if filename.endswith('.cfg'):
            print(filename)
            cfg_to_zarr('data/'+filename)
            try:
                shutil.make_archive('data/'+filename[:-4]+'.tel.zarr', 'zip', 'data/'+filename[:-4]+'.tel.zarr')
            except:
                print('Cant compress','data/'+filename[:-4]+'.tel.zarr')
            




 #   Copyright 2019 AUI, Inc. Washington DC, USA
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import numpy as np
c = 299792458
from ._sirius_utils._direction_rotate import _calc_rotation_mats, _cs_calc_rotation_mats, _directional_cosine,  _sin_project
from ._sirius_utils._apply_primary_beam import _apply_casa_airy_pb, _apply_airy_pb, apply_casa_airy_pb, apply_airy_pb
from ._sirius_utils._math_utils import _interp_array, _rot_coord
from ._sirius_utils._array_utils import _find_angle_indx, _find_val_indx
import matplotlib.pyplot as plt
import time
from numba import jit
import numba
from ._sirius_utils._constants import map_mueler_to_pol

def calc_vis(uvw,vis_data_shape,point_source_flux,point_source_ra_dec,pointing_ra_dec,phase_center_ra_dec,antenna1,antenna2,freq_chan,beam_model_map,beam_models, parallactic_angle, pol, mueller_selection):
    """
    Simulate a interferometric visibilities.
    
    Parameters
    ----------
    point_source_flux: np.array [n_point_sources,n_time, n_chan, n_pol] (singleton: n_time, n_chan)
    point_source_ra_dec: np.array [n_time, n_point_sources, 2]          (singleton: n_time)
    pointing_ra_dec: np.array [n_time, n_ant, 2]                   (singleton: n_time, n_ant)
    phase_center_ra_dec: np.array [n_time, 2]                        (singleton: n_time)
    Returns
    -------
    vis : np.array
    """
    
    
    '''
    Singleton: Can have dimension = 1
    Warning: If mosaic, do not use n_time singleton with n_ant
    '''
    
#    rotation_parms = {}
#    rotation_parms['reproject'] = True
#    rotation_parms['common_tangent_reprojection'] = False
    
    #Function to appease Numba the terrible.
    beam_models_type0, beam_models_type1, beam_types, new_beam_model_map = beam_models_to_tuple(beam_models,beam_model_map) #Needs to be done for numba
    pol = pol_code_to_index(pol)
    

    n_time, n_baseline, n_chan, n_pol = vis_data_shape
    n_ant = len(beam_model_map)
    
    #Check all dims are either 1 or n
    f_pc_time = n_time if phase_center_ra_dec.shape[0] == 1 else 1
    f_ps_time = n_time if point_source_ra_dec.shape[0] == 1 else 1
    f_sf_time = n_time if point_source_flux.shape[1] == 1 else 1
    f_sf_chan = n_chan if point_source_flux.shape[2] == 1 else 1
    
    do_pointing = False
    if pointing_ra_dec is not None:
        do_pointing = True
        f_pt_time = n_time if phase_center_ra_dec.shape[0] == 1 else 1
        f_pt_ant =  n_ant if point_source_ra_dec.shape[1] == 1 else 1
    else:
        pointing_ra_dec = np.zeros((2,2,2))
        f_pt_time = n_time
        f_pt_ant = n_ant
        
    #print(beam_models_type0, beam_models_type1)
        
    #print("do_pointing ", do_pointing )
    vis_data = np.zeros(vis_data_shape,dtype=np.complex128)
    start = time.time()
    calc_vis_jit(vis_data, uvw,tuple(vis_data_shape),point_source_flux.astype(np.complex128),point_source_ra_dec,pointing_ra_dec,phase_center_ra_dec,antenna1,antenna2,freq_chan,beam_models_type0, beam_models_type1, beam_types, new_beam_model_map, parallactic_angle, pol, mueller_selection, f_pc_time, f_ps_time, f_sf_time, f_sf_chan, f_pt_time, f_pt_ant, do_pointing)
    
    return vis_data
    
    
#@jit(nopython=True,cache=True,nogil=True)
@jit(nopython=True,nogil=True)
def calc_vis_jit(vis_data,uvw,vis_data_shape,point_source_flux,point_source_ra_dec,pointing_ra_dec,phase_center_ra_dec,antenna1,antenna2,freq_chan,beam_models_type0, beam_models_type1, beam_types, beam_model_map, parallactic_angle, pol, mueller_selection, f_pc_time, f_ps_time, f_sf_time, f_sf_chan, f_pt_time, f_pt_ant, do_pointing):

    n_time, n_baseline, n_chan, n_pol = vis_data_shape
    n_ant = len(beam_model_map)
    #vis_data = np.zeros(vis_data_shape,dtype=np.complex)
    
    n_point_source = point_source_ra_dec.shape[1]
    
    #prev_ra_dec_in =  np.zeros((4,))
    #prev_ra_dec_out = np.zeros((4,))
    prev_ra_dec_in =  np.zeros((4,),dtype=numba.float64)
    prev_ra_dec_out = np.zeros((4,),dtype=numba.float64)
    
    #print(f_pt_time, f_pt_ant, do_pointing)
    
    
    #print('pointing_ra_dec',pointing_ra_dec.dtype,point_source_ra_dec.dtype)

    for i_time in range(n_time):
        #print("Completed time step ", i_time,"of",n_time)
        pa = parallactic_angle[i_time]
        ra_dec_in = phase_center_ra_dec[i_time//f_pc_time, :]
        #print('phase_center_ra_dec',phase_center_ra_dec)
        
        for i_baseline in range(n_baseline):
            #print(i_baseline,n_baseline)
            i_ant_1 = antenna1[i_baseline]
            i_ant_2 = antenna2[i_baseline]
            if do_pointing:
                ra_dec_in_1 = pointing_ra_dec[i_time//f_pt_time,i_ant_1//f_pt_ant,:]
                ra_dec_in_2 = pointing_ra_dec[i_time//f_pt_time,i_ant_2//f_pt_ant,:]
                 
            for i_point_source in range(n_point_source):
                #s0 = time.time()
                ra_dec_out = point_source_ra_dec[i_time//f_ps_time,i_point_source,:]
                #print('ra_dec_out',ra_dec_out)
                if not(np.array_equal(prev_ra_dec_in, ra_dec_in) and np.array_equal(prev_ra_dec_out, ra_dec_out)):
                    uvw_rotmat, lmn_rot = _calc_rotation_mats(ra_dec_in, ra_dec_out)
                    #lm_sin = _sin_project(ra_dec_in,ra_dec_out)
                    lm_sin = lmn_rot #use 37 in apply_primary_beam r = 2.0*np.arcsin(np.sqrt(np.sum(lmn**2))/2.0)
                    sep = 2.0*np.arcsin(np.sqrt(np.sum(lm_sin**2))/2.0)
                    
                if do_pointing:
                    if not(np.array_equal(prev_ra_dec_in, ra_dec_in_1) and np.array_equal(prev_ra_dec_out, ra_dec_out)):
                        lm_sin_1 = _sin_project(ra_dec_in_1, ra_dec_out) # NBNBNB might need to change to _calc_rotation_mats depending
                        sep_1 = 2.0*np.arcsin(np.sqrt(np.sum(lm_sin_1**2))/2.0)
                    if not(np.array_equal(prev_ra_dec_in, ra_dec_in_2) and np.array_equal(prev_ra_dec_out, ra_dec_out)):
                        lm_sin_2 = _sin_project(ra_dec_in_2, ra_dec_out)
                        sep_2 = 2.0*np.arcsin(np.sqrt(np.sum(lm_sin_2**2))/2.0)
                
                
                
                #rad_to_arcmin = (60*180)/np.pi
                #print('ra_dec_in_1', ra_dec_in, ' ra_dec_out', ra_dec_out)
                #print(' lmn_rot ', lmn_rot, ' lm_sin ', lm_sin)
                #print(' lmn_rot ', np.sqrt(np.sum(lmn_rot**2))*rad_to_arcmin, ' lm_sin ', np.sqrt(np.sum(lm_sin**2))*rad_to_arcmin)
                #print(' lmn_rot ', 2.0*np.arcsin(np.sqrt(np.sum(lmn_rot**2))/2.0)*rad_to_arcmin, ' lm_sin ', 2.0*np.arcsin(np.sqrt(np.sum(lm_sin**2))/2.0)*rad_to_arcmin)
                #print('*'*30)
                #phase = 2*1j*np.pi*lmn_rot@(uvw[i_time,i_baseline,:]@uvw_rotmat)
                phase = 2*np.pi*lmn_rot@(uvw[i_time,i_baseline,:]@uvw_rotmat)
                
                prev_ra_dec_in = ra_dec_in
                prev_ra_dec_out = ra_dec_out
                #print("s0",time.time()-s0)
                
                #print('lmn_rot',lmn_rot)
                ##################### Apply primary beam to flux #####################
                for i_chan in range(n_chan):
                
                    
                    #s1 = time.time()
                    flux = point_source_flux[i_point_source,i_time//f_sf_time, i_chan//f_sf_chan, :]
                    
                    #print("s1",time.time()-s1)
                    
                    ################### Apply primary beams #####################
                    bm1_indx = beam_model_map[i_ant_1]
                    bm2_indx = beam_model_map[i_ant_2]
                    bm1_type = beam_types[i_ant_1]
                    bm2_type = beam_types[i_ant_2]

            

                    #print('flux_scaled',flux_scaled)
                    #s2 = time.time()
                    if do_pointing:
                        flux_scaled, outside_beam = calc_pb_scale(flux,sep_1,sep_2,bm1_indx,bm2_indx,bm1_type,bm2_type,lm_sin_1,lm_sin_2,beam_models_type0,beam_models_type1,pa,freq_chan[i_chan],mueller_selection,do_pointing)
                    else:
                        flux_scaled, outside_beam = calc_pb_scale(flux,sep,sep,bm1_indx,bm2_indx,bm1_type,bm2_type,lm_sin,lm_sin,beam_models_type0,beam_models_type1,pa,freq_chan[i_chan],mueller_selection,do_pointing)
           
                    #print("s2",time.time()-s2)
                    if not outside_beam:
                        #s3 = time.time()
                        phase_scaled = 1j*phase*freq_chan[i_chan]/c
                        #print(flux_scaled[pol])
                        #vis_data[i_time,i_baseline,i_chan,:] = vis_data[i_time,i_baseline,i_chan,:] + flux_scaled[pol]*np.exp(phase_scaled)/(1-lmn_rot[2])
                        #print("s3",time.time()-s3)
                        
                        #print(phase_scaled,type(phase),type(phase_scaled))
                        
                        for i_pol in range(n_pol):
                            #vis_data[i_time,i_baseline,i_chan,i_pol] = vis_data[i_time,i_baseline,i_chan,i_pol] + flux_scaled[pol[i_pol]]*np.exp(phase_scaled)/(1-lmn_rot[2])
                            
                            vis_data[i_time,i_baseline,i_chan,i_pol] = vis_data[i_time,i_baseline,i_chan,i_pol] + flux_scaled[pol[i_pol]]*np.exp(phase_scaled)/(1-lmn_rot[2])
                            #print(i_time,i_baseline,i_chan,i_pol,vis_data[i_time,i_baseline,i_chan,i_pol],flux_scaled[i_pol])
                            
                            #print(pb_scale*flux,np.abs(np.exp(phase_scaled)))
                    #exstract_arrays_from_bm_xds(x=7)
            #return vis_data
    
#sample_J_analytic(lmn, freq, dish_diameter, blockage_diameter, ipower, pb_func)
#sample_J(bm_J, bm_pa, bm_chan, lmn, freq, pa, delta_l, delta_m)
    

def beam_models_to_tuple(beam_models,beam_model_map):
    new_beam_model_map = np.zeros(beam_model_map.shape,dtype=np.int)
    beam_types = np.zeros(beam_model_map.shape,dtype=np.int)
    beam_models_list_type0 = []
    beam_models_list_type1 = []
    
    i_t0 = 0
    i_t1 = 0
    for i,bm in enumerate(beam_models):
        if "J" in bm:
            beam_models_list_type0.append(exstract_arrays_from_bm_xds(bm))
            beam_types[beam_model_map==i] = 0
            new_beam_model_map[beam_model_map==i] = i_t0
            i_t0 = i_t0 + 1
        else:
            beam_models_list_type1.append(exstract_vals_from_analytic_dict(bm))
            beam_types[beam_model_map==i] = 1
            new_beam_model_map[beam_model_map==i] = i_t1
            i_t1 = i_t1 + 1
            
    if not bool(beam_models_list_type0):
        d_arr = np.array([0])
        d_J = np.zeros((1,1,1,1,1),np.complex128)
        beam_models_list_type0.append((0,d_J,d_arr,d_arr,d_arr,d_arr,d_arr,0.0))
    
    if not bool(beam_models_list_type1):
        beam_models_list_type1.append((1,"none",0,0,0.0))
    
    return tuple(beam_models_list_type0), tuple(beam_models_list_type1), tuple(beam_types), tuple(new_beam_model_map)


def exstract_arrays_from_bm_xds(bm):
    bm_J = bm.J.isel(model=0).values #make sure that there is only one model
    print("bm_J", bm_J.shape)
    pa = bm.pa.values
    chan = bm.chan.values
    pol = bm.pol.values
    delta_l = bm.l[1].values - bm.l[0].values
    delta_m = bm.m[1].values - bm.m[0].values
    print(bm)
    max_rad_1GHz = bm.attrs['max_rad_1GHz']
    return (0,bm_J,pa,chan,pol,delta_l,delta_m,max_rad_1GHz)
    
def exstract_vals_from_analytic_dict(bm):
    pb_func = bm['pb_func']
    #dish_diameter = bm['dish_diameter']
    #blockage_diameter = bm['blockage_diameter']
    dish_diameter = bm['dish_diam']
    blockage_diameter = bm['blockage_diam']
    max_rad_1GHz = bm['max_rad_1GHz']
    return (1,pb_func,dish_diameter,blockage_diameter,max_rad_1GHz)


#@jit(nopython=True,cache=True,nogil=True)
@jit(nopython=True,nogil=True)
def calc_pb_scale(flux, sep1, sep2, bm1_indx,bm2_indx,bm1_type,bm2_type,lmn1,lmn2,beam_models_type0,beam_models_type1,pa,freq,mueller_selection,do_pointing):

    #print(sep1,sep2,bm1_indx,bm2_indx,bm1_type,bm2_type)
    outside_beam = False
    if (bm1_indx == bm2_indx) and (bm1_type == bm2_type) and ~do_pointing: #Antennas are the same
        if bm1_type == 0: #Checks if it is a zernike model
            bm1 = beam_models_type0[bm1_indx]
            max_rad = bm1[7]*freq/10**9 # scale max_rad_1GHz to freq
            
            if sep1 < max_rad:
                #bm1 = beam_models_type0[bm1_indx]
                J_sampled = sample_J(bm1[1],bm1[2],bm1[3],bm1[4],bm1[5],bm1[6],pa,freq,lmn1)[:,0]
                #J_sampled = np.zeros((4,),dtype=numba.complex128)
                M = make_mueler_mat(J_sampled, J_sampled, mueller_selection)
                #Add check if J sampled is < 0 and then skip this
                flux_scaled = np.dot(M,flux)
            else:
                outside_beam = True
        else:
            #bm1 = beam_models_type1[bm1_indx]
            bm1 = beam_models_type1[bm1_indx]
            max_rad = bm1[4]*freq/10**9 # scale max_rad_1GHz to freq
            if sep1 < max_rad:
                J_sampled = sample_J_analytic(bm1[1],bm1[2],bm1[3],lmn1,freq,1)
                flux_scaled = flux*J_sampled**2
            else:
                outside_beam = True
    else:
        if bm1_type == 0:
            bm1 = beam_models_type0[bm1_indx]
            max_rad = bm1[7]*freq/10**9 # scale max_rad_1GHz to freq
            if sep1 < max_rad:
                J_sampled1 = sample_J(bm1[1],bm1[2],bm1[3],bm1[4],bm1[5],bm1[6],pa,freq,lmn1)[:,0]
            else:
                outside_beam = True
        else:
            bm1 = beam_models_type1[bm1_indx]
            max_rad = bm1[4]*freq/10**9 # scale max_rad_1GHz to freq
            if sep1 < max_rad:
                J_sampled1 = sample_J_analytic(bm1[1],bm1[2],bm1[3],lmn1,freq,1)
            else:
                outside_beam = True
        if bm2_type == 0:
            bm2 = beam_models_type0[bm2_indx]
            max_rad = bm2[7]*freq/10**9 # scale max_rad_1GHz to freq
            if sep2 < max_rad:
                J_sampled2 = sample_J(bm2[1],bm2[2],bm2[3],bm2[4],bm2[5],bm2[6],pa,freq,lmn2)[:,0]
            else:
                outside_beam =  True
        else:
            bm2 = beam_models_type1[bm2_indx]
            max_rad = bm2[4]*freq/10**9 # scale max_rad_1GHz to freq
            if sep2 < max_rad:
                J_sampled2 = sample_J_analytic(bm2[1],bm2[2],bm2[3],lmn2,freq,1)
            else:
                outside_beam = True
                
        if not outside_beam:
            M = make_mueler_mat(J_sampled1, J_sampled2, mueller_selection)
            flux_scaled = np.dot(M,flux)
    if outside_beam:
        flux_scaled = np.zeros(4, dtype = numba.complex128)

    return flux_scaled, outside_beam

def pol_code_to_index(pol):
    if pol[0] in [5,6,7,8]:
        return pol-5
    if pol[0] in [9,10,11,12]:
        return pol-9
    assert False, "Unsupported pol " + str(pol)
    
#@jit(nopython=True,cache=True,nogil=True)
@jit(nopython=True,nogil=True)
def sample_J_analytic(pb_func, dish_diameter,blockage_diameter, lmn, freq, ipower):
    #pb_parms = bm
    #pb_parms['ipower'] = 1
    
    if pb_func == 'casa_airy':
        J_sampled = _apply_casa_airy_pb(lmn,freq,dish_diameter, blockage_diameter, ipower)
    elif pb_func == 'airy':
        J_sampled = _apply_airy_pb(lmn,freq,dish_diameter, blockage_diameter, ipower)
    else:
        J_sampled = 1
    
    J_sampled_array = np.zeros(4, dtype = numba.float64)
    J_sampled_array[0] = J_sampled
    J_sampled_array[3] = J_sampled
    #J_sampled = np.array([J_sampled,0,0,J_sampled])
    return J_sampled_array.astype(numba.complex128)
    
@jit(nopython=True,cache=True,nogil=True)
def sample_J(bm_J,bm_pa,bm_chan, bm_pol, bm_delta_l,bm_delta_m,pa,freq,lmn):
    pa_indx = _find_angle_indx(bm_pa,pa)
    chan_indx = _find_val_indx(bm_chan, freq)
    bm_J_sub = bm_J[pa_indx, chan_indx]

    x_rot, y_rot  = _rot_coord(lmn[0],lmn[1],pa-bm_pa[pa_indx])
    xrot = np.ones(1, dtype = numba.float64)
    xrot[0] = x_rot
    yrot = np.ones(1, dtype = numba.float64)
    yrot[0] = y_rot
    #print((xrot/bm[4]) + len(bm_J_sub[0, :, 0])//2)
    #print((yrot/bm[5]) + len(bm_J_sub[0, 0, :])//2)
    
    J_temp = _interp_array(bm_J_sub, xrot, yrot, bm_delta_l, bm_delta_m)
    
    J = np.zeros((4,J_temp.shape[1]),dtype=numba.complex128) + 1
    for i,p in enumerate(bm_pol):
        J[p] = J_temp[i,:]
        
    return J

#Non-numba versions:
def sample_J_analytic_og(bm,lmn,freq):
    pb_parms = bm
    pb_parms['ipower'] = 1
    
    if pb_parms['pb_func'] == 'casa_airy':
        J_sampled = apply_casa_airy_pb(lmn,freq,pb_parms)
    elif pb_parms['pb_func'] == 'airy':
        J_sampled = apply_airy_pb(lmn,freq,pb_parms)
    else:
        J_sampled = 1
    J_sampled = np.array([J_sampled,0,0,J_sampled])
    return J_sampled


def sample_J_og(bm,lmn,freq,pa):
# sample_J(J,pa_,chan_,lmn,chan,pa)
    bm_sub = bm
    if len(bm.pa) > 1:
        pa_indx= _find_angle_indx(bm_sub.pa.values,pa)
        bm_sub = bm_sub.isel(pa=pa_indx)
    else:
        bm_sub = bm_sub.isel(pa=0)
    if len(bm.chan) > 1:
        bm_sub = bm.J.interp(chan=freq,method='nearest')
    else:
        bm_sub = bm_sub.isel(chan=0)
        
    #print('pa values',bm_sub)
    x_rot, y_rot  = _rot_coord(lmn[0],lmn[1],pa-bm_sub.pa.values)
    
#    print(bm_sub)
#    print(lmn,pa,bm_sub.pa.values )
#    print(x_rot,y_rot)
#    print(bm_sub.J.isel(model=0).interp(l=x_rot,m=y_rot,method='linear'))
#    print(bm_sub.J.interp(l=x_rot,m=y_rot,method='linear').values)
    
    return bm_sub.J.interp(l=x_rot,m=y_rot,method='linear').values[0]

@jit(nopython=True,cache=True,nogil=True)
def make_mueler_mat(J1, J2, mueller_selection):

    #M = np.zeros((4,4),dtype=np.complex)
    M = np.zeros((4,4),dtype=numba.complex128)
    
    for m_flat_indx in mueller_selection:
        #print(m_flat_indx//4,m_flat_indx - 4*(m_flat_indx//4))
        #print(np.where(map_mueler_to_pol[m_flat_indx][0] == pol)[0][0])
        #print(pol, map_mueler_to_pol[m_flat_indx][0])
        #M[m_flat_indx//4,m_flat_indx - 4*(m_flat_indx//4)] = J1[np.where(map_mueler_to_pol[m_flat_indx][0] == pol)[0][0]]*np.conj(J2[np.where(map_mueler_to_pol[m_flat_indx][1] == pol)[0][0]])
        
        M[m_flat_indx//4,m_flat_indx - 4*(m_flat_indx//4)] = J1[map_mueler_to_pol[m_flat_indx,0]]*np.conj(J2[map_mueler_to_pol[m_flat_indx,1]])
            
    return M
    
    #map_mueler_to_pol = np.array([[0,0],[0,1],[1,0],[1,1],[0,2],[0,3],[1,2],[1,3],[2,0],[2,1],[3,0],[3,1],[2,2],[2,3],[3,2],[3,3]])

'''
    #print('pol',pol)
    if inv:
        map_mueler_to_pol = np.array([[3, 3],[3, 2],[2, 3],[2, 2],[3, 1],[3, 0],[2, 1],[2, 0],[1, 3],[1, 2],[0, 3],[0, 2],[1, 1],[1, 0],[0, 1],[0, 0]]) # np.flip(map_mueler_to_pol,axis=0)
        #map_mueler_to_pol = np.array([ [[3, 3],[3, 2],[2, 3],[2, 2]],[[3, 1],[3, 0],[2, 1],[2, 0]],[[1, 3],[1, 2],[0, 3],[0, 2]],[[1, 1],[1, 0],[0, 1],[0, 0]]])
    else:
        map_mueler_to_pol = np.array([[0,0],[0,1],[1,0],[1,1],[0,2],[0,3],[1,2],[1,3],[2,0],[2,1],[3,0],[3,1],[2,2],[2,3],[3,2],[3,3]])
        #map_mueler_to_pol = np.array([[[0,0],[0,1],[1,0],[1,1]],[[0,2],[0,3],[1,2],[1,3]],[[2,0],[2,1],[3,0],[3,1]],[[2,2],[2,3],[3,2],[3,3]]])
        
'''
    
    
    

    




 #   Copyright 2019 AUI, Inc. Washington DC, USA
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import numpy as np

#https://colab.research.google.com/drive/13kdPUQW8AD1amPzKnCqwLsA03kdI3MrP?ts=61001ba6
#https://safe.nrao.edu/wiki/pub/ALMA/SimulatorCookbook/corruptguide.pdf
#https://casadocs.readthedocs.io/en/latest/api/tt/casatools.simulator.html#casatools.simulator.simulator.setnoise
# SimACohCalc
# https://casaguides.nrao.edu/index.php/Simulating_ngVLA_Data-CASA5.4.1
# https://casaguides.nrao.edu/index.php/Corrupting_Simulated_Data_(Simulator_Tool)

def calc_a_noise(vis,uvw,beam_model_map,eval_beam_models, antenna1, antenna2, noise_parms):
    """
    Add noise to visibilities.
    
    Parameters
    ----------
    vis : np.array
    Returns
    -------
    vis : np.array
    """
    from ._sirius_utils._constants import k_B
    noise = np.zeros(vis.shape,dtype=np.complex)
    
    n_time, n_baseline, n_chan, n_pol = vis.shape
    dish_sizes = get_dish_sizes(eval_beam_models)
    
    #For now tau (Zenith Atmospheric Opacity) will be set to 0 (don't have to do elevation calculation)
    factor = (4*np.sqrt(2)*k_B*(10**23))/(noise_parms['ant_efficiency']*noise_parms['corr_efficiency']*np.pi)
    t_sys = noise_parms['t_receiver'] + noise_parms['t_atmos']*(1-noise_parms['spill_efficiency']) + noise_parms['t_cmb']
    
    del_nu = noise_parms['freq_resolution'] #should it be the total bandwidth of the spectral window?
    del_t = noise_parms['time_delta']
    
    dish_size_per_ant = dish_sizes[beam_model_map] # n_ant array with the size of each dish in meter.
    baseline_dish_diam_product = dish_size_per_ant[antenna1]*dish_size_per_ant[antenna2] # n_baseline array of dish_i*dish_j.
    
    #ms v2 weight and sigma do not have a spectral component (if we move to ms v3 this will change)
    sigma = np.tile(factor*t_sys/(baseline_dish_diam_product*np.sqrt(del_nu*del_t))[None,:], (n_time,1))
    #print('sigma[0,0]',sigma[0,0])
    #sigma[sigma < 10**-9] = 10**-9 in SimACohCalc
    weight = 1.0/(sigma**2)
    
    if not noise_parms['auto_corr']:
        sigma_full_dim = np.tile(sigma[:,:,None,None],(1,1,n_chan,n_pol))
        noise_re = np.random.normal(loc=0.0,scale=sigma_full_dim)
        noise_im = np.random.normal(loc=0.0,scale=sigma_full_dim)
        vis = vis + noise_re + 1j*noise_im
    else:
        #Most probaly will have to include the autocorrelation weight.
        auto_corr_mask = ((uvw[:,:,0]!=0) & (uvw[:,:,1]!=0)).astype(int)
        auto_corr_scale = np.copy(auto_corr_mask)
        auto_corr_scale[auto_corr_scale==0] = np.sqrt(2)

        sigma_full_dim = np.tile(sigma[:,:,None,None],(1,1,n_chan,n_pol))
        noise_re = np.random.normal(loc=0.0,scale=sigma_full_dim*auto_corr_scale[:,:,None,None])
        noise_im = np.random.normal(loc=0.0,scale=sigma_full_dim)
        
        vis = vis + noise_re + 1j*noise_im*auto_corr_mask[:,:,None,None]
    
    return vis, np.tile(weight[:,:,None],(1,1,n_pol)), np.tile(sigma[:,:,None],(1,1,n_pol))
    
    
def get_dish_sizes(eval_beam_models):
    dish_sizes = []
    for bm in eval_beam_models:
        if "J" in bm:
            dish_sizes.append(bm.attrs['dish_diam'])
        else:
            dish_sizes.append(bm['dish_diam'])
   
        
    return np.array(dish_sizes)








from setuptools import setup, find_packages

with open('README.md', "r") as fid:   #encoding='utf-8'
    long_description = fid.read()

setup(
    name='sirius',
    version='0.0.10',
    description='Simulation of Radio Interferometry from Unique Sources',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='National Radio Astronomy Observatory',
    author_email='casa-feedback@nrao.edu',
    url='https://github.com/casangi/sirius',
    license='Apache-2.0',
    packages=find_packages(),
    include_package_data=True,
    install_requires=['bokeh>=2.2.3',
                      'numba-scipy>=0.3.0',
                      'dask>=2.13.0',
                      'distributed>=2.9.3',
                      'graphviz>=0.13.2',
                      'matplotlib>=3.1.2',
                      'numba>=0.51.0',
                      'numcodecs>=0.6.3',
                      'numpy>=1.18.1',
                      'pandas>=0.25.2',
                      'scipy>=1.4.1',
                      'scikit-learn>=0.22.2',
                      'toolz>=0.10.0',
                      'xarray>=0.16.1',
                      'zarr>=2.3.2',
                      'fsspec>=0.6.2',
                      'gdown>=3.12.2',
                      'pytest>=6.2.4',
                      'ipympl>=0.7.0',
                      'dask-ms>=0.2.6',
                      'casatools>=6.3.0.48',
                      'casatasks>=6.3.0.48',
                      'casadata>=2021.8.23',
                      'python-casacore>=3.4.0',
                      'astropy>=4.3.1'
                      ],
    extras_require={
        'dev': [
            'pytest>=5.3.5',
            'black>=19.10.b0',
            'flake8>=3.7.9',
            'isort>=4.3.21',
            's3fs>=0.4.0',
            'pylint>=2.4.4',
            'radio-telescope-delay-model>=0.0.3',
            #'pytest-pep8',
            #'pytest-cov'
        ]
    }

)



#   Copyright 2019 AUI, Inc. Washington DC, USA
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import numpy as np
c = 299792458
import matplotlib.pyplot as plt
import scipy.constants
from scipy.special import j1, jn
from numba import jit
import numba_scipy.special

#@jit(nopython=True,cache=True,nogil=True)
@jit(nopython=True,nogil=True)
def _apply_casa_airy_pb(lmn,freq_chan,dish_diameter, blockage_diameter, ipower):
    #print('lmn is',lmn)
    
    if (lmn[0] != 0) and (lmn[1] != 0):
        #dish_diameter = pb_parms['dish_diameter']
        #blockage_diameter = pb_parms['blockage_diameter']
        #ipower = pb_parms['ipower']
        
        c = scipy.constants.c #299792458
        k = (2*np.pi*freq_chan)/c
        
        aperture = dish_diameter/2
        #r = np.sqrt(lmn[0]**2 + lmn[1]**2)*k*aperture
        r = 2.0*np.arcsin(np.sqrt(np.sum(lmn**2))/2.0)*k*aperture # use lmn_rot in calc vis
        #print('r',r)
        
        if blockage_diameter==0.0:
            return (2.0*j1(r)/r)**ipower
        else:
            area_ratio = (dish_diameter/blockage_diameter)**2
            length_ratio = (dish_diameter/blockage_diameter)
            return ((area_ratio * 2.0 * j1(r)/r   - 2.0 * j1(r * length_ratio)/(r * length_ratio) )/(area_ratio - 1.0))**ipower
    else:
        return 1

#@jit(nopython=True,cache=True,nogil=True)
@jit(nopython=True,nogil=True)
def _apply_airy_pb(lmn,freq_chan,dish_diameter, blockage_diameter, ipower):
    #print('lmn is',lmn)
    
    if (lmn[0] != 0) and (lmn[1] != 0):
        #dish_diameter = pb_parms['dish_diameter']
        #blockage_diameter = pb_parms['blockage_diameter']
        #ipower = pb_parms['ipower']
        
        c = scipy.constants.c #299792458
        k = (2*np.pi*freq_chan)/c
        
        aperture = dish_diameter/2
        r = np.sqrt(lmn[0]**2 + lmn[1]**2)*k*aperture
        
        if blockage_diameter==0.0:
            return (2.0*j1(r)/r)**ipower
        else:
            e = blockage_diameter/dish_diameter
            return (( 2.0 * j1(r)/r   - 2.0 * e * j1(r * e)/r )/(1.0 - e**2))**ipower
            #Changed r_grid to r ^^^
    else:
        return 1
    
    
    
#Non-jitted version:
def apply_casa_airy_pb(lmn,freq_chan,pb_parms):
    #print('lmn is',lmn)
    
    if (lmn[0] != 0) and (lmn[1] != 0):
        dish_diameter = pb_parms['dish_diameter']
        blockage_diameter = pb_parms['blockage_diameter']
        ipower = pb_parms['ipower']
        
        c = scipy.constants.c #299792458
        k = (2*np.pi*freq_chan)/c
        
        aperture = dish_diameter/2
        r = np.sqrt(lmn[0]**2 + lmn[1]**2)*k*aperture
        
        if blockage_diameter==0.0:
            return (2.0*jn(1,r)/r)**ipower
        else:
            area_ratio = (dish_diameter/blockage_diameter)**2
            length_ratio = (dish_diameter/blockage_diameter)
            return ((area_ratio * 2.0 * jn(1,r)/r   - 2.0 * jn(1, r * length_ratio)/(r * length_ratio) )/(area_ratio - 1.0))**ipower
    else:
        return 1


def apply_airy_pb(lmn,freq_chan,pb_parms):
    #print('lmn is',lmn)
    
    if (lmn[0] != 0) and (lmn[1] != 0):
        dish_diameter = pb_parms['dish_diameter']
        blockage_diameter = pb_parms['blockage_diameter']
        ipower = pb_parms['ipower']
        
        c = scipy.constants.c #299792458
        k = (2*np.pi*freq_chan)/c
        
        aperture = dish_diameter/2
        r = np.sqrt(lmn[0]**2 + lmn[1]**2)*k*aperture
        
        if blockage_diameter==0.0:
            return (2.0*jn(1,r)/r)**ipower
        else:
            e = blockage_diameter/dish_diameter
            return (( 2.0 * jn(1,r)/r   - 2.0 * e * jn(1, r * e)/r )/(1.0 - e**2))**ipower
    else:
        return 1
